import Component from '@glimmer/component';
import { tracked } from '@glimmer/tracking';
import moment from 'moment';
import { bisector } from "d3-array";
import { throttle } from '@ember/runloop';

export default class MultiScrubberComponent extends Component {
  get bisectDate() {
    return bisector(d => d.at).left;
  }

  @tracked fadeState = "opacity-0";
  @tracked currentMousePos;
  @tracked lineTextValues;
  @tracked cornerTextValues;
  @tracked line;
  @tracked movingTextContainer;
  @tracked movingTexts = [];

  get strokeWidth() {
    return this.args.width ? this.args.width : '3';
  }

  get lineColor() {
    return this.args.color ? this.args.color : 'gray';
  }

  get lineHeight() {
    return this.args.chart.getBoundingClientRect().height;
  }

  get lineTextUnits() {
    return this.args.lineTextUnits ? this.args.lineTextUnits : '';
  }

  get cornerTextUnits() {
    return this.args.cornerTextUnits ? this.args.cornerTextUnits : '';
  }

  get lineTextKey() {
    return this.args.lineTextKey ? this.args.lineTextKey : 'value'
  }

  get cornerTextKey() {
    return this.args.cornerTextKey ? this.args.cornerTextKey : 'at'
  }

  get singleValueCornerText() {
    return this.args.singleValueCornerText ? this.args.singleValueCornerText : false
  }

  get primaryDataset() {
    return this.args.datasets ? this.args.datasets[0] : []
  }

  get boxYLocation() {
    return this.args.boxYLocation ? this.args.boxYLocation : 40
  }

  get boxSize() {
    return this.args.boxSize ? this.args.boxSize : 50;
  }

  get textLineHeight() {
    return this.args.textLineHeight ? this.args.textLineHeight : this.boxSize / 3
  }
  
  get isMultiScale() {
    return typeof this.args.xScale === 'object'
  }

  get lineTexts() {
    if (!this.lineTextValues) return []
    return this.lineTextValues.map((value) => {
      if( value == null ) return ""
      if(value instanceof Date) {
        return this.formatDate(value);
      } else {
        return `${parseFloat(value.toFixed(2))}${this.lineTextUnits}`;
      }
    })
  }

  get hasLineText() {
    return this.lineTextValues && this.lineTextValues.any((textValue) => textValue || textValue === 0)
  }

  get cornerTexts() {
    if (!this.cornerTextValues) return []
    return this.cornerTextValues.map((value) => {
      if(value instanceof Date) {
        return this.formatDate(value);
      } else {
        return `${value}${this.cornerTextUnits}`;
      }
    })
  }

  updateLine = (e) => {
    const mousePos = this.mouseOrTouchPosition(e);
    this.currentMousePos = mousePos;
    if( mousePos > 0.5 ) {
      const nearestPoints = this.findNearestDatapoints(mousePos)
      if (!isNaN(mousePos)) {
        this.currentMousePos = mousePos;
        this.lineTextValues = nearestPoints.map((d) => d.y == null ? d.y : d.y[this.lineTextKey]);
        this.cornerTextValues =  this.singleValueCornerText
          ? [ nearestPoints[0].x ]
          : nearestPoints.map((d) => d.x);
      }
    }
  }

  findNearestDatapoints(mousePos) {
    if (!this.args.datasets) return [{x: "", y: ""}];

    return this.args.datasets.map((dataset, index) => {
      const xScale = this.isMultiScale ? this.args.xScale[index] : this.args.xScale
      const x0 = xScale.invert(mousePos); // x position within the element.
      return {
        x: x0,
        y: this.nearestDataPoint(dataset, x0)
      }
    });
  }

  nearestDataPoint(dataset, x0) {
    const i = this.bisectDate(dataset, x0, 1);
    const d0 = dataset[i - 1];
    const d1 = dataset[i];
    // out of dataset bounds
    if(dataset.length === 0 || x0 < dataset[0].at || x0 > dataset[dataset.length - 1].at) return
    return (d1 && d0) ? (x0 - d0.at > d1.at - x0 ? d1 : d0) : 0;
  }

  mouseOrTouchPosition(event) {
    const clientX = event.type.includes("touch") ? event.targetTouches[0].clientX : event.clientX;
    return clientX - this.args.chart.getBoundingClientRect().left;
  }

  formatDate(date) {
    const momentDate = moment(date);
    const currentYear = momentDate.year() === moment().year();

    return moment(date).format(`MMM D ${currentYear ? '' : 'YYYY'}, h:mm a`);
  }

  mousemove = (e) => {
    throttle(this, this.updateLine, e, 20);
  }

  mouseout = () => {
    this.fadeState = "opacity-0";
  }

  mouseover = (e) => {
    if(e.type === "touchstart") {
      this.currentMousePos = this.mouseOrTouchPosition(e);
      this.setMovingElementsPosition(this.currentMousePos);
    }
    this.fadeState = "opacity-100";
  }

  initElement = (element, [elementName]) => {
    if (elementName === "movingTexts") {
      this.movingTexts.push(element);
    } else {
      this[elementName] = element;
    }
  }

  animateScrubbing = () => {
    const x = (this.currentMousePos || this.currentMousePos === 0)
            ? this.getAnimatedLocation(this.line, 'x1')
            : 0

    if(x) {
      this.setMovingElementsPosition(x);
    }
  }

  setMovingElementsPosition = (x) => {
    this.line.setAttribute('x1', x)
    this.line.setAttribute('x2', x)
    this.movingTextContainer.setAttribute('x', x)
    this.movingTexts.forEach((element) => {
      element.setAttribute('x', x)
    })
  }

  getAnimatedLocation(element, attrName) {
    const currentX = parseInt(element.getAttribute(attrName)); // Get current position
    const totalDx = this.currentMousePos - currentX; // Find total x distance from cursor to element
    const acceleration = 4; // acceleration scalar
    const frameDx = totalDx / acceleration; // delta x for this frame

    return Math.abs(frameDx) > 1                      // if the delta x is less than 1 it needs to be
            ? currentX + frameDx                      // rounded or else it will register as 0px and get stuck
            : totalDx < 1 ? this.currentMousePos : currentX + 1; // note only happens really close to cursor as it deccelerates
  }
}