import Component from '@glimmer/component';
import { tracked } from '@glimmer/tracking';
import { scaleTime, scaleLinear } from "d3-scale";

export default class PlotComponent extends Component {
  @tracked dimensions;
  @tracked elementWidth;

  get heightRatio() {
    return this.args.heightRatio ? this.args.heightRatio : 0.5;
  }

  get height() {
    return (this.elementWidth ? this.elementWidth : 0) * this.heightRatio;
  }

  get transform() {
    return this.args.transform ? this.args.transform : "translate(0,0)";
  }

  get hasData() {
    return this.args.data && this.args.data.length > 0;
  }

  get isMultiDataset() {
    return this.hasData && this.args.data[0].constructor === Array;
  }

  get isMultiRange() {
    return this.args.timeRange && this.args.timeRange[0].constructor === Array;
  }

  get matchMultiRangeXscale() {
    return this.args.matchMultiRangeXscale ? this.args.matchMultiRangeXscale : true;
  }

  get yScale() {
    return this.yScaleRange.domain(this.yRange());
  }

  get xScale() {
    return this.isMultiRange
      ? this.sortedTimeRange.map(range => this.xScaleRange.domain(range))
      : this.xScaleRange.domain(this.sortedTimeRange);
  }

  get sortedTimeRange() {
    return this.args.timeRange
      ? this.isMultiRange
        ? this.adjustRanges(this.args.timeRange.map(range => range.sort(this.sortDates)))
        : this.args.timeRange.sort(this.sortDates)
      : [new Date(), new Date()]
  }

  adjustRanges(ranges) {
    if(ranges && this.isMultiRange && this.matchMultiRangeXscale) {
      const rangeLengths = ranges.map(range => range[1] - range[0])
      const maxRange = Math.max(...rangeLengths)
      return ranges.map((range) => {
        const additionalScale = maxRange - (range[1] - range[0])
        return [ range[0], new Date(range[1].getTime() + additionalScale) ]
      })
    } else {
      return ranges;
    }
  }

  sortDates(d1, d2) {
    if (d1 > d2) return 1;
    if (d1 < d2) return -1;
    return 0;
  }

  get xKey() {
    return this.args.xKey ? this.args.xKey : 'at';
  }

  get yKey() {
    return this.args.yKey ? this.args.yKey : 'value';
  }

  get xScaleRange() {
    return this.dimensions
    ? scaleTime().range([0, this.dimensions.width])
    : {}
  }

  get yScaleRange() {
    return this.dimensions
    ? scaleLinear().range([this.dimensions.height, 0])
      : {}
  }

  get filteredData() {
    if (!this.hasData) return
    if (this.isMultiDataset && this.isMultiRange) {
      return this.sortedTimeRange.map((range, index) => this.filterData(this.args.data[index], range))
    } else if (this.isMultiDataset) {
      return this.args.data.map(dataset => this.filterData(dataset));
    } else if (this.isMultiRange) {
      return this.sortedTimeRange.map(range => this.filterData(this.args.data, range))
    } else {
      return this.filterData(this.args.data);
    }
  }

  filterData(dataset, timeRange) {
    const range = timeRange ? timeRange : this.sortedTimeRange
    return dataset.filter((dataPoint) => {
      return dataPoint[this.xKey] > range[0] && dataPoint[this.xKey] < range[1];
    })
  }

  // Amount of space to buffer the min data (if set to zero, lowest datapoint will be on x axis)
  get dataBottomPadding() {
    return this.args.dataBottomPadding ? this.args.dataBottomPadding : 0
  }

  // Can easily make this scale with selected timescale.... but dont see the Y scale changing much
  // and keeping it consistent thru transitions seems like it would make a more intuitive graph
  // particularly with no visible axis'
  yRange() {
    if(!this.args.data) return [0,0]
    const y = this.yKey;
    const dataset = this.isMultiDataset ? this.args.data.flat(1) : this.args.data
    const max = Math.max(...dataset.map(d => d[y]))
    return [
      0,
      max * 1.05
    ]
  }

  resize = () => {
    const chart = this.dimensions.svg;
    this.elementWidth = chart.getBoundingClientRect().width;
    this.dimensions = { width: this.elementWidth, height: this.height, svg: chart };
    if(this.args.updateDimensions) this.args.updateDimensions(this.dimensions);
  }

  initializeChart = (el) => {
    this.elementWidth = el.getBoundingClientRect().width;
    this.dimensions = {width: this.elementWidth, height: this.height, svg: el};
    if(this.args.updateDimensions) this.args.updateDimensions( this.dimensions );
  }
}